ENVIRON_AUTH_TOKEN  = 'ZOHO_CRM_AUTH_TOKEN'
BASE_URL            = "https://crm.zoho."
API_PATH            = "/crm/private/json"
SCOPE               = 'crmapi'
MAX_PAGE_SIZE       = 200
REQUESTS_PER_SECOND = None

from .accounts.get_account import get_account  # noqa
from .contacts.get_contact import get_contact  # noqa
from .leads.get_lead import get_lead  # noqa
from .potentials.get_potential import get_potential  # noqa
from .get_record import get_record  # noqa

from .accounts.filter_accounts import filter_accounts  # noqa
from .contacts.filter_contacts import filter_contacts  # noqa
from .leads.filter_leads import filter_leads  # noqa
from .potentials.filter_potentials import filter_potentials  # noqa
from .filter_records import filter_records  # noqa

from .accounts.insert_account import insert_account  # noqa
from .contacts.insert_contact import insert_contact  # noqa
from .leads.insert_lead import insert_lead  # noqa
from .potentials.insert_potential import insert_potential  # noqa
from .insert_record import insert_record  # noqa

from .accounts.update_account import update_account  # noqa
from .contacts.update_contacts import update_contact  # noqa
from .leads.update_lead import update_lead  # noqa
from .potentials.update_potential import update_potential  # noqa
from .update_record import update_record  # noqa

from .accounts.delete_account import delete_account  # noqa
# from .contacts.delete_contacts import delete_contact  # noqa
from .leads.delete_lead import delete_lead  # noqa
# from .potentials.delete_potential import delete_potential  # noqa
from .delete_record import delete_record  # noqa



from talkzoho.regions import US
from talkzoho.projects.get_record import get_record

from talkzoho.projects.accounts import MODULE


async def get_account(*,
                      id,
                      portal_id,
                      auth_token=None,
                      region=US,
                      columns=None):
    return await get_record(
        id=id,
        module=MODULE,
        portal_id=portal_id,
        auth_token=auth_token,
        region=region,
        columns=columns)



from asyncio import Lock
from threading import Timer


class RateLimiter():

    def __init__(self, *, requests_per_minute):
        self.requests      = 0
        self.limit         = requests_per_minute
        self.requests_lock = Lock()
        self.limit_lock    = Lock()

    async def __aenter__(self):
        with await self.requests_lock:
            self.requests += 1
            hit_limit     = self.requests >= self.limit

        if hit_limit:
            await self.limit_lock.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        Timer(60.0, self.remove_request).start()

    async def remove_request(self):
        with await self.requests_lock:
            self.requests -= 1
            self.limit_lock.release()



import os

from urllib.parse import urlencode

from fuzzywuzzy import fuzz

from tornado.httpclient import AsyncHTTPClient
from tornado.web import HTTPError
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.support import BASE_URL, API_PATH, ENVIRON_AUTH_TOKEN
from talkzoho.support.utils import select_columns, unwrap_items
from talkzoho.support.accounts import MODULE


async def get_account(*,
                      auth_token=None,
                      region=US,
                      columns=None,
                      portal,
                      department,
                      id):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/' + MODULE + '/getrecordsbyid'
    endpoint = create_url(BASE_URL, tld=region, path=path)
    query    = {
        'id': id,
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN),
        'portal': portal,
        'department': department}

    if columns:
        query['selectfields'] = select_columns(MODULE, columns)

    url      = endpoint + '?' + urlencode(query)
    print(url)
    response = await client.fetch(url, method='GET')
    body     = json_decode(response.body.decode("utf-8"))

    return unwrap_items(body, single_item=True)



from talkzoho.regions import US
from talkzoho.crm.insert_record import insert_record
from talkzoho.crm.contacts import MODULE


async def insert_contact(record,
                         *,
                         auth_token=None,
                         region=US):
    return await insert_record(
        MODULE,
        record=record,
        auth_token=auth_token,
        region=region)



import os

from typing import Optional

from urllib.parse import urlencode

from fuzzywuzzy import fuzz

from tornado.httpclient import AsyncHTTPClient
from tornado.web import HTTPError
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.crm import BASE_URL, API_PATH, SCOPE, MAX_PAGE_SIZE, ENVIRON_AUTH_TOKEN
from talkzoho.crm.utils import select_columns, unwrap_items


async def filter_records(module: str,
                         *,
                         auth_token: Optional[str]=None,
                         term: Optional[str]=None,
                         region: str=US,
                         columns: Optional[list]=None,
                         offset: int=0,
                         limit: Optional[int]=None):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/' + module + '/getRecords'
    endpoint = create_url(BASE_URL, tld=region, path=path)

    if limit == 0:
        return []
    elif not term and limit and limit <= MAX_PAGE_SIZE:
        batch_size = limit
    else:
        batch_size = MAX_PAGE_SIZE

    paging     = True
    from_index = offset + 1  # Zoho indexes at one not zero
    to_index   = offset + batch_size
    results    = []

    # Loop until we reach index we need, unless their is a search term.
    # If search term we need all records.
    while paging and (term or limit is None or to_index <= limit):
        query = {
            'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN),
            'fromIndex': from_index,
            'toIndex': to_index,
            'newFormat': 2,
            'version': 2,
            'scope': SCOPE}

        if columns:
            query['selectColumns'] = select_columns(module, columns)

        url      = endpoint + '?' + urlencode(query)
        response = await client.fetch(url, method='GET')
        body     = json_decode(response.body.decode('utf-8'))

        try:
            items = unwrap_items(body)
        except HTTPError as http_error:
            # if paging and hit end suppress error
            # unless first request caused the 404
            if http_error.status_code == 404 and from_index - 1 != offset:
                paging = False
            else:
                raise
        else:
            results   += items
            from_index = to_index + 1
            to_index  += batch_size

    def fuzzy_score(resource):
        values = [str(v) for v in resource.values() if v]
        target = ' '.join(values)
        return fuzz.partial_ratio(term, target)

    if term:
        results = sorted(results, key=fuzzy_score, reverse=True)

    return results[:limit]



from talkzoho.regions import US
from talkzoho.crm.insert_record import insert_record
from talkzoho.crm.potentials import MODULE


async def insert_potential(record,
                           *,
                           auth_token=None,
                           region=US):
    return await insert_record(
        MODULE,
        record=record,
        auth_token=auth_token,
        region=region)



from talkzoho.regions import US
from talkzoho.crm.update_record import update_record
from talkzoho.crm.potentials import MODULE, PRIMARY_KEY


async def update_potential(record,
                           *,
                           auth_token=None,
                           region=US):
    return await update_record(
        MODULE,
        primary_key=PRIMARY_KEY,
        record=record,
        auth_token=auth_token,
        region=region)



from talkzoho.regions import US
from talkzoho.crm.filter_records import filter_records
from talkzoho.crm.potentials import MODULE


async def filter_potentials(*,
                            auth_token=None,
                            term=None,
                            region=US,
                            columns=None,
                            offset=0,
                            limit=None):
    return await filter_records(
        MODULE,
        auth_token=auth_token,
        term=term,
        region=region,
        columns=columns,
        offset=offset,
        limit=limit)



from tornado.web import HTTPError


def select_columns(resource, columns):
    return resource.lower() + '(' + ','.join(columns) + ')' if columns else ''


def to_zoho_value(value):
    if value is None:
        value == ''
    else:
        return value


def record_to_xml_data(record: dict):
    lines  = ['<FL val="{}"><![CDATA[{}]]></FL>'.format(k, to_zoho_value(v))
              for k, v in record.items()]

    return ''.join(lines)


def wrap_items(items, *, module_name: str):
    if type(items) is not list:
        items = [items]

    rows = ['<row no="{}">{}</row>'.format(index + 1, record_to_xml_data(item))
            for index, item in enumerate(items)]

    return '<{module_name}>{rows}</{module_name}>'.format(
        module_name=module_name,
        rows=''.join(rows))


def unwrap_items(response, single_item=False):
    try:
        result   = response['response']['result']

        if len(result) == 1:
            # Don't know the resource name but should be the only key
            resource = list(result.values())[0]
            rows     = resource['row']
        elif len(result) == 2:
            # On update message returns two keys message and record
            single_item = True
            rows        = result['recorddetail']
        else:
            raise ValueError('Unexpected looking response.')

        # wrap single resource results in array
        items = rows if isinstance(rows, list) else [rows]

        if single_item and len(items) != 1:
            ValueError('More then one resource was returned.')

        items = [translate_item(i) for i in items]
        return items[0] if single_item else items
    except (AssertionError, KeyError):
        return unwrap_error(response)


def unwrap_error(zoho_error):
    try:
        response   = zoho_error['response']
        filtered   = {key: value for key, value in response.items() if key.lower() != 'uri'}  # noqa

        # Dont know the error name but should ony be one key left
        assert len(filtered) == 1
        _, error      = filtered.popitem()
        code, message = error['code'], error['message']

        # 5000 code returned on successful deletion
        if code == '5000':
            return True

        status_code = http_status_code(zoho_code=code)
        raise HTTPError(status_code, reason=message)
    except (AssertionError, KeyError, IndexError):
        raise ValueError("Couldn't parse zoho result")


def http_status_code(*, zoho_code):  # pragma: no cover
    zoho_code = str(zoho_code)

    if zoho_code in ["4000", "4401", "4600", "4831", "4832", "4835", "4101", "4420"]:  # noqa
        return 400  # bad request
    elif zoho_code in ["4501", "4834"]:
        return 401  # unauthorised
    elif zoho_code in ["4502", "4890"]:
        return 402  # payment required
    elif zoho_code in ["4487", "4001", "401", "401.1", "401.2", "401.3"]:
        return 403  # forbidden
    elif zoho_code in ["4102", "4103", "4422"]:
        return 404  # not found
    elif zoho_code in []:
        return 405  # method not allowed
    elif zoho_code in ["4807"]:
        return 413  # payload too large
    elif zoho_code in ["4424"]:
        return 415  # payload too large
    elif zoho_code in ["4101", "4809"]:
        return 423  # locked
    elif zoho_code in ["4820", "4421", "4423"]:
        return 429  # too many requests
    else:
        return 500  # internal server error


def translate_item(item):
    fields = item.get('fl', item.get('FL'))
    fields = fields if isinstance(fields, list) else [fields]

    def nullify(value):
        return None if value == 'null' else value

    return {kwarg['val']: nullify(kwarg['content']) for kwarg in fields}



from tornado.ioloop import IOLoop
from functools import partial


def create_url(base_url, *, tld, path):
    return base_url + tld + path


def wait(f, *args, **kwargs):
    loaded_f = partial(f, *args, **kwargs)
    return IOLoop.current().run_sync(loaded_f)



from tornado.web import HTTPError


def unwrap_items(response, *, single_item=False, columns=None):
    """
    Project response items always comeback in a list
    even when they are GETs for an ID. Single Item tells
    this unwrapper that you're expecting a single item out.
    """
    try:
        # Dont know the resource name but should be the only key
        assert len(response) == 1
        resources = list(response.values())[0]

        if single_item and len(resources) != 1:
            ValueError('More then one resource was returned.')

        resources = [
            {k: v for k, v in r.items() if not columns or k in columns}
            for r in resources]

        return resources[0] if single_item else resources
    except (AssertionError, KeyError):
        unwrap_error(response)


def unwrap_error(zoho_error):
    try:
        # Dont know the error name but should ony be one key left
        code        = zoho_error['error']['code']
        message     = zoho_error['error']['message']
        status_code = http_status_code(zoho_code=code)

        raise HTTPError(status_code, reason=message)
    except (AssertionError, KeyError, IndexError):
        raise ValueError("Couldn't parse zoho result")


def http_status_code(*, zoho_code):  # pragma: no cover
    zoho_code = str(zoho_code)

    if zoho_code in [6401, 6891, 6403, 6831, 6832, 6500]:  # noqa
        return 400  # bad request
    elif zoho_code in [6401, 6890]:
        return 401  # unauthorised
    elif zoho_code in [6504, 6404]:
        return 404  # not found
    else:
        return 500  # internal server error



from talkzoho.regions import US
from talkzoho.crm.delete_record import delete_record
from talkzoho.crm.leads import MODULE


async def delete_lead(*,
                      auth_token=None,
                      region=US,
                      id):
    return await delete_record(
        MODULE,
        auth_token=auth_token,
        region=region,
        id=id)



import os

from typing import Optional, Union

from urllib.parse import urlencode

from tornado.httpclient import AsyncHTTPClient
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.crm import BASE_URL, API_PATH, SCOPE, ENVIRON_AUTH_TOKEN
from talkzoho.crm.utils import select_columns, unwrap_items


async def get_record(module: str,
                     *,
                     auth_token: Optional[str]=None,
                     region: str=US,
                     columns: Optional[list]=None,
                     id: Union[int, str]):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/' + module + '/getRecordById'
    endpoint = create_url(BASE_URL, tld=region, path=path)
    query    = {
        'id': id,
        'scope': SCOPE,
        'version': 2,
        'newFormat': 2,
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN)}

    if columns:
        query['selectColumns'] = select_columns(module, columns)

    url      = endpoint + '?' + urlencode(query)
    response = await client.fetch(url, method='GET')
    body     = json_decode(response.body.decode('utf-8'))

    return unwrap_items(body, single_item=True)



import os

from urllib.parse import urlencode

from tornado.httpclient import AsyncHTTPClient
from tornado.escape import json_decode

from talkzoho.utils import create_url
from talkzoho.regions import US
from talkzoho.projects import BASE_URL, API_PATH, ENVIRON_AUTH_TOKEN
from talkzoho.projects.utils import unwrap_items


async def get_record(*,
                     id,
                     module,
                     portal_id,
                     auth_token=None,
                     region=US,
                     columns=None):
    query = urlencode({
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN)})

    client   = AsyncHTTPClient()
    path     = API_PATH + '/portal/' + portal_id + '/' + module + '/' + id + '/'
    endpoint = create_url(BASE_URL, tld=region, path=path)

    url      = endpoint + '?' + query
    response = await client.fetch(url, method='GET')
    body     = json_decode(response.body.decode("utf-8"))

    return unwrap_items(body, single_item=True, columns=columns)



ENVIRON_AUTH_TOKEN  = 'ZOHO_BOOKS_AUTH_TOKEN'
BASE_URL            = "https://books.zoho."
API_PATH            = "/api/v3"
SCOPE               = 'booksapi'
MAX_PAGE_SIZE       = 200
REQUESTS_PER_SECOND = 2.5


from .price_books.get_price_book import get_price_book  # noqa
from .price_books.filter_price_books import filter_price_books  # noqa

from .items.filter_items import filter_items  # noqa



from talkzoho.regions import US
from talkzoho.crm.update_record import update_record
from talkzoho.crm.accounts import MODULE, PRIMARY_KEY


async def update_account(record,
                         *,
                         auth_token=None,
                         region=US):
    return await update_record(
        MODULE,
        primary_key=PRIMARY_KEY,
        record=record,
        auth_token=auth_token,
        region=region)



import os

from urllib.parse import urlencode

from fuzzywuzzy import fuzz

from tornado.httpclient import AsyncHTTPClient
from tornado.escape import json_decode
from tornado.web import HTTPError

from talkzoho.utils import create_url
from talkzoho.regions import US
from talkzoho.projects import BASE_URL, API_PATH, ENVIRON_AUTH_TOKEN, MAX_PAGE_SIZE  # noqa
from talkzoho.projects.utils import unwrap_items


async def filter_projects(*,
                          auth_token=None,
                          term=None,
                          region=US,
                          columns=None,
                          offset=0,
                          limit=None,
                          portal_id):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/portal/' + portal_id + '/projects/'
    endpoint = create_url(BASE_URL, tld=region, path=path)

    if limit == 0:
        return []
    elif not term and limit and limit <= MAX_PAGE_SIZE:
        batch_size = limit
    else:
        batch_size = MAX_PAGE_SIZE

    paging     = True
    from_index = offset
    to_index   = offset + batch_size - 1
    results    = []

    # Loop until we reach index we need, unless their is a search term.
    # If search term we need all records.
    while paging:
        query = urlencode({
            'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN),
            'index': from_index + 1,  # Zoho indexes at one not zero,
            'range': batch_size})

        url           = endpoint + '?' + query
        response      = await client.fetch(url, method='GET')
        reached_limit = limit and to_index + 1 >= limit

        if response.code == 204 or (term is None and reached_limit):
            break

        body  = json_decode(response.body.decode("utf-8"))
        items = unwrap_items(body, columns=columns)

        results   += items
        from_index = to_index + 1
        to_index  += batch_size

    def fuzzy_score(resource):
        values = [str(v).lower() for v in resource.values() if v]
        target = ' '.join(values)
        return fuzz.partial_ratio(term, target)

    if term:
        results = sorted(results, key=fuzzy_score, reverse=True)

    return results[:limit]



ENVIRON_AUTH_TOKEN  = 'ZOHO_PROJECTS_AUTH_TOKEN'
BASE_URL            = "https://projectsapi.zoho."
API_PATH            = "/restapi"
MAX_PAGE_SIZE       = 100
REQUESTS_PER_SECOND = None  # TODO: done per endpoint at 100 in two minutes


from .accounts.get_account import get_account  # noqa

from .projects.get_project import get_project  # noqa
from .projects.filter_projects import filter_projects  # noqa



ENVIRON_AUTH_TOKEN  = 'ZOHO_SUPPORT_AUTH_TOKEN'
BASE_URL            = "https://support.zoho."
API_PATH            = "/api/json"
MAX_PAGE_SIZE       = 200
REQUESTS_PER_SECOND = None  # TODO: done per endpoint at 100 in two minutes


from .accounts.get_account import get_account  # noqa
from .accounts.filter_accounts import filter_accounts  # noqa



from talkzoho.regions import US

from talkzoho.books.filter_module import filter_module

RESOURCE = 'items'


async def filter_items(*,
                       auth_token=None,
                       organization_id=None,
                       term=None,
                       region=US,
                       columns=None,
                       offset=0,
                       limit=None):
    return await filter_module(
        RESOURCE,
        auth_token=auth_token,
        organization_id=organization_id,
        term=term,
        region=region,
        columns=columns,
        offset=offset,
        limit=limit)



from tornado.web import HTTPError


def select_columns(resource, columns):
    return resource + '(' + ','.join(columns) + ')' if columns else ''


def unwrap_items(response, single_item=False):
    try:
        result   = response['response']['result']

        # Dont know the resource name but should be the only key
        assert len(result) == 1
        resource = list(result.values())[0]

        # wrap single resource results in array
        rows  = resource['row']
        items = rows if isinstance(rows, list) else [rows]

        if single_item and len(items) != 1:
            ValueError('More then one resource was returned.')

        items = [translate_item(i) for i in items]
        return items[0] if single_item else items
    except (AssertionError, KeyError):
        unwrap_error(response)


def unwrap_error(zoho_error):
    try:
        response   = zoho_error['response']
        filtered   = {key: value for key, value in response.items() if key.lower() != 'uri'}  # noqa

        # Dont know the error name but should ony be one key left
        assert len(filtered) == 1
        _, error      = filtered.popitem()
        code, message = error['code'], error['message']

        status_code = http_status_code(zoho_code=code)
        raise HTTPError(status_code, reason=message)
    except (AssertionError, KeyError, IndexError):
        raise ValueError("Couldn't parse zoho result")


def http_status_code(*, zoho_code):  # pragma: no cover
    zoho_code = str(zoho_code)

    if zoho_code in ["4832"]:
        return 404  # not found
    else:
        return 500  # internal server error


def translate_item(item):
    fields = item.get('fl', item.get('FL'))
    fields = fields if isinstance(fields, list) else [fields]

    def nullify(value):
        return None if value == 'null' else value

    return {kwarg['val']: nullify(kwarg['content']) for kwarg in fields}



from talkzoho.regions import US
from talkzoho.crm.get_record import get_record
from talkzoho.crm.potentials import MODULE


async def get_potential(*,
                       auth_token=None,
                       region=US,
                       columns=None,
                       id):
    return await get_record(
        MODULE,
        auth_token=auth_token,
        region=region,
        columns=columns,
        id=id)



from talkzoho.regions import US
from talkzoho.crm.get_record import get_record
from talkzoho.crm.contacts import MODULE


async def get_contact(*,
                      auth_token=None,
                      region=US,
                      columns=None,
                      id):
    return await get_record(
        MODULE,
        auth_token=auth_token,
        region=region,
        columns=columns,
        id=id)



import os

from typing import Optional

from urllib.parse import urlencode

from tornado.httpclient import AsyncHTTPClient
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.crm import BASE_URL, API_PATH, SCOPE, ENVIRON_AUTH_TOKEN
from talkzoho.crm.utils import wrap_items, unwrap_items


async def update_record(module: str,
                        record: dict,
                        *,
                        primary_key: str,
                        auth_token: Optional[str]=None,
                        region: str=US,
                        trigger_workflow: bool=True):
    client     = AsyncHTTPClient()
    path       = API_PATH + '/' + module + '/updateRecords'
    endpoint   = create_url(BASE_URL, tld=region, path=path)

    id_        = record.pop(primary_key)
    xml_record = wrap_items(record, module_name=module)

    body = urlencode({
        'scope': SCOPE,
        'version': 2,
        'newFormat': 2,
        'duplicateCheck': 1,
        'wfTrigger': str(trigger_workflow).lower(),
        'id': id_,
        'xmlData': xml_record,
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN)})

    response = await client.fetch(endpoint, method='POST', body=body)
    body     = json_decode(response.body.decode('utf-8'))

    return unwrap_items(body, single_item=True)['Id']



from talkzoho.regions import US
from talkzoho.crm.filter_records import filter_records
from talkzoho.crm.contacts import MODULE


async def filter_contacts(*,
                          auth_token=None,
                          term=None,
                          region=US,
                          columns=None,
                          offset=0,
                          limit=None):
    return await filter_records(
        MODULE,
        auth_token=auth_token,
        term=term,
        region=region,
        columns=columns,
        offset=offset,
        limit=limit)



import os

from urllib.parse import urlencode

from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPError as HTTPClientError
from tornado.web import HTTPError
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.books import BASE_URL, API_PATH, SCOPE, ENVIRON_AUTH_TOKEN

RESOURCE = 'pricebooks'


async def get_price_book(*,
                         auth_token=None,
                         region=US,
                         columns=None,
                         organization_id,
                         id):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/' + RESOURCE + '/' + str(id)
    endpoint = create_url(BASE_URL, tld=region, path=path)
    url      = endpoint + '?' + urlencode({
        'organization_id': organization_id,
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN)})

    try:
        response = await client.fetch(url, method='GET')
    except HTTPClientError as http_error:
        http_code = http_error.code
        response  = json_decode(http_error.response.body.decode("utf-8"))
        message   = str(response['code']) + ': ' + response['message']
        raise HTTPError(http_code, reason=message)
    else:
        response = json_decode(response.body.decode("utf-8"))
        results  = [v for k, v in response.items() if k not in ['code', 'message']]

        if len(results) != 1:
            ValueError('More then one resource was returned.')

        return results[0]



from setuptools import setup
from setuptools import find_packages


setup(
    name='talkzoho',
    version='0.9.0',
    description='Asynchronous wrapper for Zoho\'s numerous APIs',
    url='https://github.com/A2Z-Cloud/Talk-Zoho',
    packages=find_packages(exclude=('tests', 'tests.*')),
    install_requires=[
        'fuzzywuzzy',
        'python-Levenshtein',
        'requests',
        'tornado'])



import os

import requests

from talkzoho.regions import US
from talkzoho.utils import create_url
from talkzoho.crm import BASE_URL, API_PATH, SCOPE, ENVIRON_AUTH_TOKEN


def upload_file(module,
                *,
                auth_token=None,
                region=US,
                id,
                file_name,
                file_bytes):
    parameters   = {
        'id': int(id),
        'scope': SCOPE,
        'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN)}
    files = {
        'content': (file_name, file_bytes)}

    path     = API_PATH + '/' + module + '/uploadFile'
    endpoint = create_url(BASE_URL, tld=region, path=path)

    # TODO: replace with tornado client
    response = requests.post(endpoint, data=parameters, files=files)
    response.raise_for_status()

    return True



import os
from math import ceil

from urllib.parse import urlencode

from fuzzywuzzy import fuzz

from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPError as HTTPClientError
from tornado.web import HTTPError
from tornado.escape import json_decode

from talkzoho.regions import US
from talkzoho.utils import create_url

from talkzoho.books import BASE_URL, API_PATH, SCOPE, MAX_PAGE_SIZE, ENVIRON_AUTH_TOKEN


async def filter_module(module,
                        *,
                        auth_token=None,
                        organization_id=None,
                        term=None,
                        region=US,
                        columns=None,
                        offset=0,
                        limit=None):
    client   = AsyncHTTPClient()
    path     = API_PATH + '/' + module
    endpoint = create_url(BASE_URL, tld=region, path=path)

    if limit == 0:
        return []
    elif not term and limit and limit <= MAX_PAGE_SIZE:
        batch_size = limit
    else:
        batch_size = MAX_PAGE_SIZE

    paging     = True
    page_index = max(ceil(offset / batch_size), 1)
    results    = []

    # Loop until we reach index we need, unless their is a search term.
    # If search term we need all records.
    while paging and (term or not limit or len(results) < limit):
        url = endpoint + '?' + urlencode({
            'scope': SCOPE,
            'authtoken': auth_token or os.getenv(ENVIRON_AUTH_TOKEN),
            'organization_id': organization_id,
            'per_page': batch_size,
            'page': page_index})

        try:
            response = await client.fetch(url, method='GET')
        except HTTPClientError as http_error:
            http_code = http_error.code
            response  = json_decode(http_error.response.body.decode("utf-8"))
            message   = str(response['code']) + ': ' + response['message']
            raise HTTPError(http_code, reason=message)
        else:
            response    = json_decode(response.body.decode("utf-8"))
            results    += response[module]
            page_index += 1
            paging      = response['page_context']['has_more_page']

    def fuzzy_score(price_list):
        values = [str(v).lower() for v in price_list.values() if v]
        target = ' '.join(values)
        return fuzz.partial_ratio(term, target)

    if term:
        results = sorted(results, key=fuzzy_score, reverse=True)

    results = results[:limit]
    if columns:
        return [{k: pl[k] for k in columns if k in columns} for pl in results]
    else:
        return results
